import pandas as pd
from pandas._testing import assert_frame_equal

import datacode.hooks as dc_hooks
from datacode import DataSource

from tests.pipeline.base import PipelineTest

COUNTER = 0


def increase_counter_hook(*args, **kwargs):
    global COUNTER
    COUNTER += 1
    return kwargs


def raise_error(*args, **kwargs):
    raise ValueError


def increase_counter_hook_discard_first_arg(*args):
    increase_counter_hook()
    return args[1:]


def increase_counter_hook_return_only_second_arg(*args):
    increase_counter_hook()
    return args[1]


def increment_df_a_variable_hook(source: DataSource, df: pd.DataFrame) -> pd.DataFrame:
    df['a'] += 1
    return df


class HooksTest(PipelineTest):

    def teardown_method(self, *args, **kwargs):
        super().teardown_method(*args, **kwargs)
        dc_hooks.reset_hooks()


class TestPipelineHooks(HooksTest):

    def test_no_pipeline_hook(self):
        counter_value = COUNTER
        dgp = self.create_generator_pipeline()
        dgp.execute()
        assert COUNTER == counter_value

    def test_reset_hook(self):
        counter_value = COUNTER
        dc_hooks.on_begin_execute_pipeline = increase_counter_hook
        dc_hooks.reset_hooks()
        dgp = self.create_generator_pipeline()
        dgp.execute()
        assert COUNTER == counter_value

    def test_chain_at_end(self):
        counter_value = COUNTER
        dc_hooks.on_begin_execute_pipeline = increase_counter_hook
        dc_hooks.chain('on_begin_execute_pipeline', raise_error)
        dgp = self.create_generator_pipeline()
        with self.assertRaises(ValueError) as cm:
            dgp.execute()
        assert COUNTER == counter_value + 1

    def test_chain_at_beginning(self):
        counter_value = COUNTER
        dc_hooks.on_begin_execute_pipeline = increase_counter_hook
        dc_hooks.chain('on_begin_execute_pipeline', raise_error, at_end=False)
        dgp = self.create_generator_pipeline()
        with self.assertRaises(ValueError) as cm:
            dgp.execute()
        assert COUNTER == counter_value

    def test_on_begin_execute_pipeline(self):
        counter_value = COUNTER
        dc_hooks.on_begin_execute_pipeline = increase_counter_hook
        dgp = self.create_generator_pipeline()
        dgp.execute()
        assert COUNTER == counter_value + 1
        dc_hooks.chain('on_begin_execute_pipeline', increase_counter_hook)
        dgp.execute()
        assert COUNTER == counter_value + 3

    def test_on_end_execute_pipeline(self):
        counter_value = COUNTER
        dc_hooks.on_end_execute_pipeline = increase_counter_hook
        dgp = self.create_generator_pipeline()
        dgp.execute()
        assert COUNTER == counter_value + 1
        dc_hooks.chain('on_end_execute_pipeline', increase_counter_hook)
        dgp.execute()
        assert COUNTER == counter_value + 3

    def test_on_begin_execute_operation(self):
        counter_value = COUNTER
        dc_hooks.on_begin_execute_operation = increase_counter_hook
        dgp = self.create_generator_pipeline()
        dgp.execute()
        assert COUNTER == counter_value + 1
        dc_hooks.chain('on_begin_execute_operation', increase_counter_hook)
        dgp = self.create_generator_pipeline()
        dgp.execute()
        assert COUNTER == counter_value + 3

    def test_on_end_execute_operation(self):
        counter_value = COUNTER
        dc_hooks.on_end_execute_operation = increase_counter_hook
        dgp = self.create_generator_pipeline()
        dgp.execute()
        assert COUNTER == counter_value + 1
        dc_hooks.chain('on_end_execute_operation', increase_counter_hook)
        dgp = self.create_generator_pipeline()
        dgp.execute()
        assert COUNTER == counter_value + 3

    def test_on_begin_load_source(self):
        counter_value = COUNTER
        dc_hooks.on_begin_load_source = increase_counter_hook
        self.create_csv()
        ds = self.create_source(df=None)
        df = ds.df
        assert COUNTER == counter_value + 1
        dc_hooks.chain('on_begin_load_source', increase_counter_hook)
        ds = self.create_source(df=None)
        df = ds.df
        assert COUNTER == counter_value + 3

    def test_on_end_load_source(self):
        dc_hooks.on_end_load_source = increment_df_a_variable_hook
        self.create_csv()
        ds = self.create_source(df=None)
        df = ds.df
        assert_frame_equal(df, self.expect_df_no_rename_a_plus_1)
        dc_hooks.chain('on_end_load_source', increment_df_a_variable_hook)
        ds = self.create_source(df=None)
        df = ds.df
        assert_frame_equal(df, self.expect_df_no_rename_a_plus_2)

    def test_on_begin_apply_variable_transform(self):
        counter_value = COUNTER
        dc_hooks.on_begin_apply_variable_transform = increase_counter_hook_discard_first_arg
        self.create_csv()
        cols = self.create_columns(transform_data='series')
        ds = self.create_source(df=None, columns=cols)
        df = ds.df
        assert COUNTER == counter_value + 2
        dc_hooks.chain('on_begin_apply_variable_transform', increase_counter_hook_discard_first_arg)
        cols = self.create_columns(transform_data='series')
        ds = self.create_source(df=None, columns=cols)
        df = ds.df
        assert COUNTER == counter_value + 6

    def test_on_end_apply_variable_transform(self):
        counter_value = COUNTER
        dc_hooks.on_end_apply_variable_transform = increase_counter_hook_return_only_second_arg
        self.create_csv()
        cols = self.create_columns(transform_data='series')
        ds = self.create_source(df=None, columns=cols)
        df = ds.df
        assert COUNTER == counter_value + 2
        dc_hooks.chain('on_end_apply_variable_transform', increase_counter_hook_return_only_second_arg)
        cols = self.create_columns(transform_data='series')
        ds = self.create_source(df=None, columns=cols)
        df = ds.df
        assert COUNTER == counter_value + 6

    def test_on_begin_apply_source_transform(self):
        counter_value = COUNTER
        dc_hooks.on_begin_apply_source_transform = increase_counter_hook_return_only_second_arg
        dtp = self.create_transformation_pipeline()
        dtp.execute()
        assert COUNTER == counter_value + 1
        dc_hooks.chain('on_begin_apply_source_transform', increase_counter_hook_return_only_second_arg)
        dtp = self.create_transformation_pipeline()
        dtp.execute()
        assert COUNTER == counter_value + 3

    def test_on_end_apply_source_transform(self):
        counter_value = COUNTER
        dc_hooks.on_end_apply_source_transform = increase_counter_hook_return_only_second_arg
        dtp = self.create_transformation_pipeline()
        dtp.execute()
        assert COUNTER == counter_value + 1
        dc_hooks.chain('on_end_apply_source_transform', increase_counter_hook_return_only_second_arg)
        dtp = self.create_transformation_pipeline()
        dtp.execute()
        assert COUNTER == counter_value + 3


import uuid
from typing import Optional

from graphviz import Digraph
from mixins import EqOnAttrsMixin, EqHashMixin

from datacode.graph.base import GraphObject


class Node(GraphObject, EqHashMixin, EqOnAttrsMixin):
    _recursive_hash_convert = True
    equal_attrs = ['label', 'name', 'id', 'node_kwargs']

    def __init__(self, name: str, label: str = 'name', id_: Optional[str] = None, **node_kwargs):
        if label == 'name':
            self.label = name
        elif label is None:
            self.label = ''
        else:
            self.label = label

        self.name = name
        if id_ is None:
            self.id = str(uuid.uuid4())
        else:
            self.id = id_
        self.node_kwargs = node_kwargs

    def add_to_graph(self, graph: Digraph):
        graph.node(self.id, label=self.label, **self.node_kwargs)

    def __repr__(self):
        return f'<Node(name={self.name})>'



import unittest
from copy import deepcopy

import datacode as dc
from datacode import DataSource

ORIG_COPY_KEYS = deepcopy(DataSource.copy_keys)


class OptionsTest(unittest.TestCase):
    def tearDown(self) -> None:
        dc.options.reset()


class TestOptions(OptionsTest):
    def test_set_class_attr_and_reset(self):
        new_value = ["a", "b"]
        assert DataSource.copy_keys == ORIG_COPY_KEYS
        dc.options.set_class_attr("DataSource", "copy_keys", new_value)
        assert DataSource.copy_keys == new_value
        dc.options.reset()
        assert DataSource.copy_keys == ORIG_COPY_KEYS

    def test_set_class_attr_context_manager(self):
        new_value = ["a", "b"]
        assert DataSource.copy_keys == ORIG_COPY_KEYS
        with dc.options:
            dc.options.set_class_attr("DataSource", "copy_keys", new_value)
            assert DataSource.copy_keys == new_value
        assert DataSource.copy_keys == ORIG_COPY_KEYS
        with dc.options.set_class_attr("DataSource", "copy_keys", new_value):
            assert DataSource.copy_keys == new_value
        assert DataSource.copy_keys == ORIG_COPY_KEYS



from typing import List, Union
from unittest.mock import patch

from datacode import DataExplorer, DataSource
from datacode.models.pipeline.base import DataPipeline
from tests.pipeline.base import PipelineTest

COUNT = 0
EXPECT_DIFFICULTY = 50


def str_counter() -> str:
    global COUNT
    COUNT += 1
    return str(COUNT)


def get_list_if_df_else_get_num(data: Union[DataSource, DataPipeline]) -> Union[int, List[int]]:
    df = data.df
    if df is None:
        return 0
    return [len(df), len(df)]


class ExplorerTest(PipelineTest):

    def teardown_method(self, *args, **kwargs):
        super().teardown_method()
        global COUNT
        COUNT = 0

    def create_explorer(self):
        dp = self.create_merge_pipeline()
        ds = self.create_source()

        data = dict(
            sources=[ds],
            pipelines=[dp]
        )

        explorer = DataExplorer.from_dict(data)
        return explorer


class TestCreateExplorer(ExplorerTest):

    def test_create_from_sources_and_pipelines_dict(self):
        dp = self.create_merge_pipeline()
        ds = self.create_source()

        data = dict(
            sources=[ds],
            pipelines=[dp]
        )

        explorer = DataExplorer.from_dict(data)
        got_ds, got_dp = explorer.items
        assert ds == got_ds
        assert dp == got_dp


class TestExplorerGraph(ExplorerTest):

    @patch('uuid.uuid4', str_counter)
    def test_graph_no_attrs(self):
        explorer = self.create_explorer()
        graph_str = str(explorer.graph())
        assert graph_str.startswith('digraph "Data Explorer" {')
        assert '1 -> 4' in graph_str
        assert '2 -> 4' in graph_str
        assert '5' in graph_str
        assert '2 [label=two]' in graph_str
        assert '1 [label=one]' in graph_str
        assert graph_str.endswith("\n}")

    @patch('uuid.uuid4', str_counter)
    def test_graph_attrs(self):
        explorer = self.create_explorer()
        graph_str = str(explorer.graph(include_attrs=['location', 'operation_options']))
        assert graph_str.startswith('digraph "Data Explorer" {')
        assert '1 -> 4' in graph_str
        assert '2 -> 4' in graph_str
        assert '5 [label="{  | location = tests/generated_files/data.csv }" shape=Mrecord]' in graph_str
        assert '4 [label="{  | operation_options = [\\<DataMerge(on_names=[\'C\'], merge_function=left_merge_df, kwargs=\\{\\})\\>] }" shape=Mrecord]'
        assert '2 [label="{ two | location = tests/generated_files/data2.csv }" shape=Mrecord]' in graph_str
        assert '1 [label="{ one | location = tests/generated_files/data.csv }" shape=Mrecord]' in graph_str
        assert graph_str.endswith("\n}")

    @patch('uuid.uuid4', str_counter)
    def test_graph_function_dict(self):
        explorer = self.create_explorer()
        func_dict = dict(obs=get_list_if_df_else_get_num)
        graph_str = str(explorer.graph(func_dict=func_dict))
        assert graph_str.startswith('digraph "Data Explorer" {')
        assert '1 -> 4' in graph_str
        assert '2 -> 4' in graph_str
        assert '5 [label="{  | obs_0 = 3 | obs_1 = 3 }" shape=Mrecord]' in graph_str
        assert '4 [label="{  | obs = 0 }" shape=Mrecord]'
        assert '2 [label="{ two | obs_0 = 2 | obs_1 = 2 }" shape=Mrecord]' in graph_str
        assert '1 [label="{ one | obs_0 = 3 | obs_1 = 3 }' in graph_str
        assert graph_str.endswith("\n}")


class TestDifficulty(ExplorerTest):

    def test_get_difficulty_by_property(self):
        explorer = self.create_explorer()
        assert explorer.difficulty == EXPECT_DIFFICULTY

    def test_get_difficulty_by_class_method(self):
        dp = self.create_merge_pipeline()
        ds = self.create_source()

        difficulty = DataExplorer.get_difficulty_for([dp, ds])
        assert difficulty == EXPECT_DIFFICULTY

    def test_get_custom_difficulty(self):
        dp = self.create_merge_pipeline()
        dp.difficulty = 15
        ds = self.create_source()
        ds.difficulty = 5

        difficulty = DataExplorer.get_difficulty_for([dp, ds])
        assert difficulty == 20

    def test_no_overlapping_difficulty(self):
        dp = self.create_merge_pipeline()
        ds = self.create_source()
        da1 = self.create_analysis_pipeline(source=ds)
        da1.name = 'Analysis One'
        da2 = self.create_analysis_pipeline(source=ds)
        da2.name = 'Analysis Two'

        difficulty = DataExplorer.get_difficulty_for([dp, ds, da1, da2])
        assert difficulty == EXPECT_DIFFICULTY + 100

    def test_get_difficulty_between(self):
        ds = self.create_source(name='one', difficulty=10)
        gp = self.create_generator_pipeline()
        gp.name = 'Generate One'
        ds2 = self.create_source(pipeline=gp, name='two', difficulty=20)
        dp = self.create_merge_pipeline(data_sources=[ds, ds2])
        dp.name = 'Merge One Two'
        da1 = self.create_analysis_pipeline(source=dp)
        da1.name = 'Analysis One'
        ds3 = self.create_source(name='three', difficulty=70)
        da2 = self.create_analysis_pipeline(source=ds2)
        da2.name = 'Analysis Two'

        de = DataExplorer([ds, gp, ds2, dp, da1, da2])
        # Single on each side
        difficulty = de.difficulty_between([ds2], [da1])
        assert difficulty == 120
        difficulty = de.difficulty_between([gp], [da1])
        assert difficulty == 170
        difficulty = de.difficulty_between([ds], [dp])
        assert difficulty == 60

        # Multiple begins, single end
        difficulty = de.difficulty_between([ds, ds2], [da1])
        assert difficulty == 130

        # Multiple ends, single begin
        difficulty = de.difficulty_between([ds2], [da1, da2])
        assert difficulty == 170

        # Multiple begins, multiple ends
        difficulty = de.difficulty_between([ds, ds2], [da1, da2])
        assert difficulty == 180

        # Test errors
        with self.assertRaises(ValueError) as cm:
            difficulty = de.difficulty_between([ds3], [da1])
            exc = cm.exception
            assert 'no direct link between the items could be determined' == str(exc)
        ds4 = self.create_source(name='four')
        with self.assertRaises(ValueError) as cm:
            difficulty = de.difficulty_between([ds4], [da1])
            exc = cm.exception
            assert f'must pass items which are already in DataExplorer, but got {ds4}' == str(exc)



from datacode import AnalysisOptions
from datacode.models.logic.partial import partial
import datacode.hooks as dc_hooks
from tests.pipeline.base import PipelineTest, analysis_from_source
import tests.test_hooks as th


class TestDataAnalysisPipeline(PipelineTest):

    def test_create_and_run_analysis_pipeline_from_source(self):
        dap = self.create_analysis_pipeline()
        dap.execute()

        assert dap.result.result == self.ds_one_analysis_result
        self.assert_all_pipeline_operations_have_pipeline(dap)

    def test_analysis_pipeline_output(self):
        options = AnalysisOptions(analysis_from_source, out_path=self.csv_path_output)
        dap = self.create_analysis_pipeline(options=options)
        dap.execute()

        with open(self.csv_path_output, 'r') as f:
            result_from_file = int(f.read())

        assert result_from_file == self.ds_one_analysis_result
        self.assert_all_pipeline_operations_have_pipeline(dap)

    def test_create_and_run_analysis_pipeline_from_merge_pipeline(self):
        dmp = self.create_merge_pipeline()
        dap = self.create_analysis_pipeline(source=dmp)
        dap.execute()

        assert dap.result.result == self.ds_one_and_two_analysis_result
        self.assert_ordered_pipeline_operations(dap, [dmp, dap])

    def test_create_and_run_analysis_pipeline_from_transformation_pipeline(self):
        dtp = self.create_transformation_pipeline()
        dap = self.create_analysis_pipeline(source=dtp)
        dap.execute()

        assert dap.result.result == self.ds_one_transformed_analysis_result
        self.assert_ordered_pipeline_operations(dap, [dtp, dap])

    def test_create_and_run_analysis_pipeline_from_generation_pipeline(self):
        dgp = self.create_generator_pipeline()
        dap = self.create_analysis_pipeline(source=dgp)
        dap.execute()

        assert dap.result.result == self.ds_one_generated_analysis_result
        self.assert_ordered_pipeline_operations(dap, [dgp, dap])

    def test_create_and_run_multiple_analysis_pipelines_from_same_transformation_pipeline(self):
        dtp = self.create_transformation_pipeline()
        dap1 = self.create_analysis_pipeline(source=dtp)

        analysis_from_source_2 = partial(analysis_from_source, sum_offset=10)
        ao2 = AnalysisOptions(analysis_from_source_2)
        dap2 = self.create_analysis_pipeline(source=dtp, options=ao2)

        counter_value = th.COUNTER
        dc_hooks.on_begin_apply_source_transform = th.increase_counter_hook_return_only_second_arg

        dap1.execute()
        dap2.execute()

        dc_hooks.reset_hooks()

        assert dap1.operations[0] is dap2.operations[0]
        assert dap1.operations[0].data_source is dap2.operations[0].data_source
        assert dap1.result.result == self.ds_one_transformed_analysis_result
        assert dap2.result.result == self.ds_one_transformed_analysis_result_offset_10
        assert th.COUNTER == counter_value + 1  # transform operation called only once
        self.assert_ordered_pipeline_operations(dap1, [dtp, dap1])
        self.assert_ordered_pipeline_operations(dap2, [dtp, dap2])

    def test_create_and_run_multiple_analysis_pipelines_from_same_transformation_pipeline_with_always_rerun(self):
        dtp = self.create_transformation_pipeline(always_rerun=True)
        dap1 = self.create_analysis_pipeline(source=dtp)

        analysis_from_source_2 = partial(analysis_from_source, sum_offset=10)
        ao2 = AnalysisOptions(analysis_from_source_2)
        dap2 = self.create_analysis_pipeline(source=dtp, options=ao2)

        counter_value = th.COUNTER
        dc_hooks.on_begin_apply_source_transform = th.increase_counter_hook_return_only_second_arg

        dap1.execute()
        dap2.execute()

        dc_hooks.reset_hooks()

        assert dap1.operations[0] is dap2.operations[0]
        assert dap1.operations[0].data_source is dap2.operations[0].data_source
        assert dap1.result.result == self.ds_one_transformed_analysis_result
        assert dap2.result.result == self.ds_one_transformed_analysis_result_offset_10
        assert th.COUNTER == counter_value + 2  # transform operation called twice as always rerun
        self.assert_ordered_pipeline_operations(dap1, [dtp, dap1])
        self.assert_ordered_pipeline_operations(dap2, [dtp, dap2])


from typing import Any, Tuple, Dict


class DatacodeOptions:
    """
    Allows setting options for the datacode library

    :Usage:

    Use as a context manager with a single change:

    >>> with dc.options.set_class_attr("DataSource", "copy_keys", ['a', 'b']):
    >>>     # Do something
    >>> # Now options have been reset

    Usage as a context manager with multiple changes:

    >>> with dc.options:
    >>>     dc.options.set_class_attr("DataSource", "copy_keys", ['a', 'b'])
    >>>     # More options changes, then desired operations

    Plain usage:

    >>> dc.options.set_class_attr("DataSource", "copy_keys", ['a', 'b'])
    >>> # Now change lasts, until (optionally) calling
    >>> dc.options.reset()
    """

    _orig_class_attrs: Dict[Tuple[str, str], Any] = {}

    def reset(self):
        """
        Undo any changes made through the options interface
        :return:
        """
        for (class_name, attr), orig_value in self._orig_class_attrs.items():
            self._set_class_attr(class_name, attr, orig_value)
        self._orig_class_attrs = {}

    def set_class_attr(
        self, class_name: str, attr: str, value: Any
    ) -> "DatacodeOptions":
        """
        Sets an attribute on a datacode class

        :param class_name: Name of a class in the main datacode namespace
        :param attr: Attribute to be updated on the class
        :param value: Value to set the attribute to
        :return: same options instance
        """
        orig_value = self._set_class_attr(class_name, attr, value)
        self._orig_class_attrs[(class_name, attr)] = orig_value
        return self

    def _set_class_attr(self, class_name: str, attr: str, value: Any) -> Any:
        import datacode as dc

        klass = getattr(dc, class_name)
        orig_value = getattr(klass, attr)
        setattr(klass, attr, value)
        return orig_value

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.reset()


options = DatacodeOptions()



import itertools
from typing import Sequence, Union, List, Optional, Dict, Tuple

from mixins import ReprMixin

from datacode.models.source import DataSource
from datacode.models.pipeline.base import DataPipeline
from datacode.graph.base import GraphObject, Graphable, GraphFunction


class DataExplorer(Graphable, ReprMixin):
    """
    Pass it DataSources and DataPipelines to explore and inspect them
    """

    repr_cols = ["item_names"]

    def __init__(
        self,
        items: Sequence[Union[DataSource, DataPipeline]],
        name: str = "Data Explorer",
    ):
        self.items = items
        self.name = name
        super().__init__()

    @classmethod
    def from_dict(cls, data: dict) -> "DataExplorer":
        items = _get_list_of_items_from_nested_dict(data)
        return cls(items)

    @property
    def item_names(self) -> List[str]:
        return [item.name for item in self.items]

    def __getitem__(self, item):
        return self.items[item]

    @property
    def difficulty(self) -> float:
        """
        The total difficulty of all items in the explorer

        :return: Total difficulty
        """
        total = 0
        for item in self:
            item_diff = getattr(item, "difficulty", 0)
            total += item_diff
        return total

    @classmethod
    def get_difficulty_for(cls, items: Sequence[Union[DataSource, DataPipeline]]):
        """
        Get total difficulty for a set of items

        :param items: Items which have difficulty
        :return: Total difficulty
        """
        de = cls(items)
        return de.difficulty

    def difficulty_between(
        self,
        begin: Sequence[Union[DataSource, DataPipeline]],
        end: Sequence[Union[DataSource, DataPipeline]],
    ) -> float:
        """
        Calculates the total difficulty to execute all the passed pipelines and
        sources, starting from begin and ending at end.

        :param begin: Items which are earlier in the pipelines
        :param end: Items which are later in the pipelines
        :return: Total difficulty

        :Notes:

            Items will not be double counted. If a pipeline is on the path
            to run between multiple begin and end items, its difficulty will
            be added only once.

        """
        for item in list(begin) + list(end):
            if item not in self:
                raise ValueError(
                    f"must pass items which are already in DataExplorer, but got {item}"
                )

        # Start from end, working back to beginning in a tree search. Once the path has been
        # found, then work back up to add up the difficulties
        total, found = _work_back_through_all_data_totaling_difficulty_until(end, begin)
        if not found:
            raise ValueError(f"no direct link between the items could be determined")
        return total

    def _graph_contents(
        self,
        include_attrs: Optional[Sequence[str]] = None,
        func_dict: Optional[Dict[str, GraphFunction]] = None,
    ) -> List[GraphObject]:
        # TODO [#94]: more efficient DataExplorer.graph
        #
        # Examining last_modified or pipeline_last_modified on
        # a large pipeline structure is extremely slow. Performance
        # of DataExplorer graphing could be improved if it first found
        # only the terminal pipelines and sources and used only those,
        # as the nested is included anyway.
        all_contents = []
        for item in self.items:
            all_contents.extend(item._graph_contents(include_attrs, func_dict))
        all_contents = list(set(all_contents))
        return all_contents


def _get_list_of_items_from_nested_dict(data: dict):
    all_items = []
    for key, value in data.items():
        if isinstance(value, dict):
            nested_items = _get_list_of_items_from_nested_dict(value)
            all_items.extend(nested_items)
        elif isinstance(value, (list, tuple)):
            all_items.extend(value)
        else:
            all_items.append(value)
    return all_items


def _work_back_through_all_data_totaling_difficulty_until(
    items: Sequence[Union[DataSource, DataPipeline]],
    end_items: Sequence[Union[DataSource, DataPipeline]],
) -> Tuple[float, bool]:
    any_found = False
    total = 0
    counted_item_ids: List[int] = []
    for item, end in itertools.product(items, end_items):
        sub_total, found = _work_back_through_data_totaling_difficulty_until(
            item, end, counted_item_ids
        )
        if found:
            any_found = True
            total += sub_total
    return total, any_found


def _work_back_through_data_totaling_difficulty_until(
    item: Union[DataSource, DataPipeline],
    end: Union[DataSource, DataPipeline],
    counted_item_ids: List[int],
) -> Tuple[float, bool]:
    if isinstance(item, DataSource):
        return _work_back_through_data_source_totaling_difficulty_until(
            item, end, counted_item_ids
        )
    elif isinstance(item, DataPipeline):
        return _work_back_through_pipeline_totaling_difficulty_until(
            item, end, counted_item_ids
        )
    else:
        raise ValueError(
            f"expected DataSource or DataPipeline, got {item} of type {type(item)}"
        )


def _work_back_through_data_source_totaling_difficulty_until(
    item: DataSource, end: Union[DataSource, DataPipeline], counted_item_ids: List[int]
) -> Tuple[float, bool]:
    total = 0
    if item.pipeline is None:
        # Hit end of this branch but did not find
        return total, False
    if item.pipeline == end:
        # Hit end of this branch and found
        total += item.pipeline.difficulty
        total += item.difficulty
        return total, True
    # There is a pipeline, and it is not the end, so continue recursively
    sub_total, sub_found = _work_back_through_data_totaling_difficulty_until(
        item.pipeline, end, counted_item_ids
    )
    return _aggregate_subtotal(item, total, sub_total, sub_found, counted_item_ids)


def _work_back_through_pipeline_totaling_difficulty_until(
    item: DataPipeline,
    end: Union[DataSource, DataPipeline],
    counted_item_ids: List[int],
) -> Tuple[float, bool]:
    total = 0
    for sub_item in item.data_sources:
        if sub_item == end:
            # Hit the end of this branch and found
            if id(item) not in counted_item_ids:
                # This item was not previously added, so add this item's difficulty
                total += item.difficulty
                counted_item_ids.append(id(item))
            if id(sub_item) not in counted_item_ids:
                # The sub-item was not previously added, so add the sub-item's difficulty
                total += sub_item.difficulty
                counted_item_ids.append(id(sub_item))
            return total, True
        sub_total, sub_found = _work_back_through_data_totaling_difficulty_until(
            sub_item, end, counted_item_ids
        )
        total, _ = _aggregate_subtotal(
            item, total, sub_total, sub_found, counted_item_ids
        )
        if sub_found:
            return total, True
    # Did not find in nested pipeline, this layer is also invalid
    return total, False


def _aggregate_subtotal(
    item: Union[DataSource, DataPipeline],
    total: float,
    sub_total: float,
    sub_found: bool,
    counted_item_ids: List[int],
):
    if sub_found:
        # now we are on the correct path
        total += sub_total
        if id(item) not in counted_item_ids:
            # This item was not previously added, so add this item's difficulty
            total += item.difficulty
            counted_item_ids.append(id(item))
        return total, True

    # Did not find in nested pipeline, this layer is also invalid
    return total, False



import datetime
import time

from pandas.testing import assert_frame_equal

from datacode import DataSource, Column
import datacode.hooks as dc_hooks
from tests.pipeline.base import PipelineTest
import tests.test_hooks as th


class TestDataTransformationPipeline(PipelineTest):

    def test_create_and_run_transformation_pipeline_from_func(self):
        dtp = self.create_transformation_pipeline()
        dtp.execute()

        assert_frame_equal(dtp.df, self.expect_func_df)
        self.assert_all_pipeline_operations_have_pipeline(dtp)

    def test_create_and_run_transformation_pipeline_from_transform(self):
        dtp = self.create_transformation_pipeline(func=self.source_transform)
        dtp.execute()

        assert_frame_equal(dtp.df, self.expect_loaded_df_with_transform)
        self.assert_all_pipeline_operations_have_pipeline(dtp)

    def test_auto_run_pipeline_by_load_source_with_no_location(self):
        dtp = self.create_transformation_pipeline()

        ds = DataSource(pipeline=dtp, location=self.csv_path_output)
        df = ds.df
        assert_frame_equal(df, self.expect_func_df)
        self.assert_all_pipeline_operations_have_pipeline(dtp)

    def test_auto_run_pipeline_by_load_source_with_no_location_and_shared_columns(self):
        self.create_csv()
        all_cols = self.create_columns()

        def transform_func(source: DataSource) -> DataSource:
            new_ds = DataSource(df=source.df, columns=all_cols)
            return new_ds

        dtp = self.create_transformation_pipeline(func=transform_func)

        ds = DataSource(pipeline=dtp, location=self.csv_path_output, columns=all_cols)
        df = ds.df
        assert_frame_equal(df, self.expect_loaded_df_rename_only)
        self.assert_all_pipeline_operations_have_pipeline(dtp)

    def test_create_nested_transformation_pipeline(self):
        dtp = self.create_transformation_pipeline()

        dtp2 = self.create_transformation_pipeline(source=dtp)
        dtp2.execute()
        assert_frame_equal(dtp2.df, self.expect_df_double_source_transform)
        self.assert_ordered_pipeline_operations(dtp2, [dtp, dtp2])

    def test_create_nested_generation_pipeline(self):
        dgp = self.create_generator_pipeline()

        dtp = self.create_transformation_pipeline(source=dgp)
        dtp.execute()

        assert_frame_equal(dtp.df, self.expect_generated_transformed)
        self.assert_ordered_pipeline_operations(dtp, [dgp, dtp])

    def test_create_nested_merge_pipeline(self):
        dmp = self.create_merge_pipeline()

        dtp = self.create_transformation_pipeline(source=dmp)
        dtp.execute()

        assert_frame_equal(dtp.df, self.expect_merged_1_2_both_transformed)
        self.assert_ordered_pipeline_operations(dtp, [dmp, dtp])

    def test_original_variables_not_affected_by_transform(self):
        self.create_csv()
        a, b, c = self.create_variables()
        ac = Column(a, 'a')
        bc = Column(b, 'b')
        cc = Column(c, 'c')
        all_cols = [ac, bc, cc]
        source = self.create_source(df=None, columns=all_cols)
        dtp = self.create_transformation_pipeline(source=source)
        dtp.execute()

        assert not a.applied_transforms
        assert not b.applied_transforms
        assert not c.applied_transforms
        self.assert_all_pipeline_operations_have_pipeline(dtp)


    def test_transform_on_source_with_normal_and_transformed_of_same_variable(self):
        self.create_csv()
        a, b, c = self.create_variables(transform_data='cell', apply_transforms=False)
        ac = Column(a, 'a')
        bc = Column(b, 'b')
        cc = Column(c, 'c')
        all_cols = [ac, bc, cc]
        include_vars = [
            a,
            a.add_one_cell(),
            b,
            c
        ]
        source = self.create_source(df=None, columns=all_cols, load_variables=include_vars)
        dtp = self.create_transformation_pipeline(source=source)
        dtp.execute()

        assert_frame_equal(dtp.df, self.expect_func_df_with_a_and_a_transformed)
        self.assert_all_pipeline_operations_have_pipeline(dtp)

    def test_nested_last_modified_of_source_greater_than_pipeline(self):
        counter_value = th.COUNTER
        dc_hooks.on_begin_apply_source_transform = th.increase_counter_hook_return_only_second_arg

        dmp = self.create_merge_pipeline()

        dtp = self.create_transformation_pipeline(source=dmp)
        self.create_csv_for_2()
        ds = self.create_source(df=None, location=self.csv_path2, pipeline=dtp)

        # Should not run pipeline as source is newer
        df = ds.df

        dc_hooks.reset_hooks()
        assert_frame_equal(df, self.test_df2)
        assert th.COUNTER == counter_value  # transform operation not called
        self.assert_ordered_pipeline_operations(dtp, [dmp, dtp])

    def test_nested_last_modified_of_source_less_than_pipeline(self):
        counter_value = th.COUNTER
        dc_hooks.on_begin_apply_source_transform = th.increase_counter_hook_return_only_second_arg

        dmp = self.create_merge_pipeline()

        dtp = self.create_transformation_pipeline(source=dmp)
        self.create_csv_for_2()
        ds = self.create_source(df=None, location=self.csv_path2, pipeline=dtp)

        time.sleep(0.01)
        self.create_csv()  # now earlier source is more recently modified
        assert dmp.data_sources[0].last_modified > ds.last_modified

        # Should run pipeline as original source is newer
        df = ds.df

        dc_hooks.reset_hooks()
        assert_frame_equal(df, self.expect_merged_1_2_both_transformed)
        assert th.COUNTER == counter_value + 1  # transform operation called once
        self.assert_ordered_pipeline_operations(dtp, [dmp, dtp])

    def test_nested_last_modified_of_source_less_than_earlier_source(self):
        counter_value = th.COUNTER
        dc_hooks.on_begin_apply_source_transform = th.increase_counter_hook_return_only_second_arg

        dtp = self.create_transformation_pipeline()
        self.create_csv_for_2()
        cols2 = self.create_columns()
        ds2 = self.create_source(df=None, location=self.csv_path2, pipeline=dtp, columns=cols2)
        dtp2 = self.create_transformation_pipeline(source=ds2)
        time.sleep(0.01)
        self.create_csv_for_3()
        ds3 = self.create_source(df=None, location=self.csv_path3, pipeline=dtp2)

        time.sleep(0.01)
        self.create_csv()
        ds1 = dtp.data_sources[0]
        # now first source is most recently modified, followed by third source. Middle source is oldest
        assert ds1.last_modified > ds2.last_modified
        assert ds3.last_modified > ds2.last_modified

        # Should run both pipelines as original source is newest
        df = ds3.df

        dc_hooks.reset_hooks()
        assert_frame_equal(df, self.expect_df_double_source_transform)
        assert th.COUNTER == counter_value + 2  # transform operation called once
        self.assert_all_pipeline_operations_have_pipeline(dtp)
        self.assert_all_pipeline_operations_have_pipeline(dtp2)



"""
Hooks to run functions during datacode operations globally

:Usage:

    >>> import datacode.hooks as dc_hooks
    ...
    >>> def log_begin_execute(pipeline: DataPipeline):
    ...     print(f"running {pipeline.name}")
    ...
    >>> dc_hooks.on_begin_execute_pipeline = log_begin_execute

:All the Hooks:

    Each hook may expect different arguments and have different return values.
    Here is the list of the hooks, along with the expected arguments and return values.

"""
from copy import deepcopy
from typing import TYPE_CHECKING, Tuple, Callable

import pandas as pd


if TYPE_CHECKING:
    from datacode.models.pipeline.base import DataPipeline
    from datacode.models.pipeline.operations.operation import DataOperation
    from datacode.models.source import DataSource
    from datacode.models.column.column import Column
    from datacode.models.variables.variable import Variable
    from datacode.models.transform.transform import Transform
    from datacode.models.transform.source import SourceTransform

_orig_locals = {
    key: value for key, value in locals().items() if not key.startswith("_")
}


def on_begin_execute_pipeline(pipeline: "DataPipeline") -> None:
    """
    Called at the beginning of :meth:`DataPipeline.execute`

    :param pipeline: The pipeline which is about to be executed
    :return: None
    """
    pass


def on_end_execute_pipeline(pipeline: "DataPipeline") -> None:
    """
    Called at the end of :meth:`DataPipeline.execute`

    :param pipeline: The pipeline which was just executed
    :return: None
    """
    pass


def on_begin_execute_operation(operation: "DataOperation") -> None:
    """
    Called at the beginning of :meth:`DataOperation.execute`

    :param operation: The operation which is about to be executed
    :return: None
    """
    pass


def on_end_execute_operation(operation: "DataOperation") -> None:
    """
    Called at the end of :meth:`DataOperation.execute`

    :param operation: The operation which was just executed
    :return: None
    """
    pass


def on_begin_load_source(source: "DataSource") -> None:
    """
    Called at the beginning of :meth:`DataSource._load`, which
    is usually called when loading :paramref:`.DataSource.df`
    from either location or pipeline.

    :param source: DataSource for which DataFrame is about to be loaded
    :return:
    """
    pass


def on_end_load_source(source: "DataSource", df: pd.DataFrame) -> pd.DataFrame:
    """
    Called at the end of :meth:`DataSource._load`, which
    is usually called when loading :paramref:`.DataSource.df`
    from either location or pipeline.

    The DataFrame returned from this function will be
    set as :paramref:`.DataSource.df`

    :param source: DataSource for which DataFrame was just loaded
    :param df:
    :return: DataFrame which will be set as :paramref:`.DataSource.df`
    """
    return df


def on_begin_apply_variable_transform(
    transform: "Transform", source: "DataSource", col: "Column", var: "Variable"
) -> Tuple["DataSource", "Column", "Variable"]:
    """
    Called at the beginning of :meth:`Transform._apply_transform_for_column_and_variable_to_source`,
    which is called while transforming variables.

    The DataSource, Column, and Variable returned from this function will be used
    to do the variable transform

    :param transform: Transform which is about to be applied
    :param source: DataSource in which the transform is about to be applied
    :param col: Column on which the transform is about to be applied
    :param var: Variable on which the transform is about to be applied
    :return: The items which will be used to transform the variable
    """
    return source, col, var


def on_end_apply_variable_transform(
    transform: "Transform", source: "DataSource", col: "Column", var: "Variable"
) -> "DataSource":
    """
    Called at the end of :meth:`Transform._apply_transform_for_column_and_variable_to_source`,
    which is called while transforming variables.

    The DataSource returned from this function will be used
    as the results of the variable transform

    :param transform: Transform which was just applied
    :param source: DataSource in which the transform was just applied
    :param col: Column on which the transform was just applied
    :param var: Variable on which the transform was just applied
    :return: The DataSource which will be used as the results
        of the variable transform
    """
    return source


def on_begin_apply_source_transform(transform: 'SourceTransform', source: 'DataSource') -> 'DataSource':
    """
    Called at the beginning of :meth:`SourceTransform.apply`, which
    is called in TransformPipeline

    The DataSource returned from this function will be used
    to do the source transform

    :param transform: SourceTransform which is about to be applied
    :param source: DataSource on which the SourceTransform will be applied
    :return: The DataSource on which to apply the source transform
    """
    return source


def on_end_apply_source_transform(transform: 'SourceTransform', source: 'DataSource') -> 'DataSource':
    """
    Called at the end of :meth:`SourceTransform.apply`, which
    is called in TransformPipeline

    The DataSource returned from this function will be used
    as the result of the source transform

    :param transform: SourceTransform which was just applied
    :param source: DataSource on which the SourceTransform was just applied
    :return: The DataSource which will become the result of
        the source transformation
    """
    return source


_new_locals = {key: value for key, value in locals().items() if not key.startswith("_")}
_hook_keys = [key for key in _new_locals if key not in _orig_locals]
_orig_hooks = deepcopy(
    {key: value for key, value in _new_locals.items() if key in _hook_keys}
)
_return_none_hook_names = [
    'on_begin_execute_pipeline',
    'on_end_execute_pipeline',
    'on_begin_execute_operation',
    'on_end_execute_operation',
    'on_begin_load_source',
]
_skip_first_arg_return_rest_hook_names = [
    'on_begin_apply_variable_transform',
]
_return_argument_2_hook_names = [
    'on_end_load_source',
    'on_end_apply_variable_transform',
    'on_begin_apply_source_transform',
    'on_end_apply_source_transform',
]


def reset_hooks() -> None:
    """
    Go back to original dummy hooks, removes all user settings of hooks

    :return: None

    :Notes:

        This function is not a hook itself.
        Instead it is a utility method meant to be called by the user.

    """
    for key, value in _orig_hooks.items():
        globals()[key] = value


def chain(hook_name: str, func: Callable, at_end: bool = True) -> None:
    """
    Add a function to a hook to be called at the end
    of the prior registered function(s), rather than
    replacing the hook entirely

    :param hook_name: Name of hook in the hooks module
    :param func: Function to be added in the call chain
        for the hook. Must have the same call signature
        as the hook.
    :param at_end: Whether to call this new function
        after the existing hook. Pass False to call this
        new function before the existing hook.
    :return:

    :Notes:

        This function is not a hook itself.
        Instead it is a utility method meant to be called by the user.

    """
    if hook_name not in _hook_keys:
        raise ValueError(f'could not find hook with name {hook_name}')

    existing_hook = globals()[hook_name]
    if at_end:
        first = existing_hook
        last = func
    else:
        first = func
        last = existing_hook

    if hook_name in _return_none_hook_names:
        def chained_hook(*args, **kwargs):
            first(*args, **kwargs)
            last(*args, **kwargs)
    elif hook_name in _return_argument_2_hook_names:
        def chained_hook(*args, **kwargs):
            arg_2 = first(*args, **kwargs)
            new_args = (args[0], arg_2)
            if len(args) > 2:
                new_args = new_args + args[1:]
            return last(*new_args, **kwargs)
    elif hook_name in _skip_first_arg_return_rest_hook_names:
        def chained_hook(*args, **kwargs):
            ret = first(*args, **kwargs)
            return last(args[0], *ret)
    else:
        raise NotImplementedError(f'hook {hook_name} has not been classified according to its return values')

    globals()[hook_name] = chained_hook


__all__ = _hook_keys + ["reset_hooks", "chain"]



import uuid
from typing import Callable, Dict, List, Sequence, Optional, TYPE_CHECKING, Union, Any

if TYPE_CHECKING:
    from datacode.graph.node import Node
    from datacode.models.source import DataSource
    from datacode.models.pipeline.base import DataPipeline

from graphviz import Digraph

GraphFunction = Callable[[Union['DataSource', 'DataPipeline']], Any]

ESCAPE_CHARS = ['<', '>', '{', '}', '|']


class GraphObject:

    def add_to_graph(self, graph: Digraph):
        raise NotImplementedError


class Graphable:
    name: str

    def __init__(self):
        self._node_id = str(uuid.uuid4())

    def _graph_contents(self, include_attrs: Optional[Sequence[str]] = None,
                        func_dict: Optional[Dict[str, GraphFunction]] = None) -> List[GraphObject]:
        raise NotImplementedError

    def graph(self, include_attrs: Optional[Sequence[str]] = None,
              func_dict: Optional[Dict[str, GraphFunction]] = None) -> Digraph:
        elems = self._graph_contents(include_attrs, func_dict)
        graph = Digraph(self.name)
        for elem in elems:
            elem.add_to_graph(graph)
        return graph

    def primary_node(self, include_attrs: Optional[Sequence[str]] = None,
                     func_dict: Optional[Dict[str, GraphFunction]] = None) -> 'Node':
        from datacode.graph.node import Node

        label_parts = [self.name if self.name is not None else '']
        label_parts.extend(self._include_attrs_labels(include_attrs))
        label_parts.extend(self._function_dict_labels(func_dict))
        label = get_multirow_label_from_parts(label_parts)
        if len(label_parts) == 1:
            # Did not find any included attributes
            return Node(self.name, id_=self._node_id)

        # Has valid included attributes
        return Node(label, shape='Mrecord', id_=self._node_id)

    def _include_attrs_labels(self, include_attrs: Optional[Sequence[str]] = None) -> List[str]:
        if include_attrs is None:
            return []

        label_parts = []
        for attr in include_attrs:
            if hasattr(self, attr):
                value = getattr(self, attr)
                str_value = get_valid_label_part(str(value))
                value_label = f'{attr} = {str_value}'
                label_parts.append(value_label)
        return label_parts

    def _function_dict_labels(self, func_dict: Optional[Dict[str, GraphFunction]] = None) -> List[str]:
        if func_dict is None:
            return []

        label_parts = []
        for attr_name, func in func_dict.items():
            label_parts.extend(self._function_labels(func, attr_name))
        return label_parts

    def _function_labels(self, func: Optional[GraphFunction] = None, attr_name: Optional[str] = None) -> List[str]:
        if func is None:
            return []

        if attr_name is None:
            attr_name = func.__name__

        orig_parts = func(self)  # type: ignore

        if not isinstance(orig_parts, (list, tuple)):
            return [f'{attr_name} = {get_valid_label_part(str(orig_parts))}']

        # Handle multiple returned items
        label_parts: List[str] = []
        for i, value in enumerate(orig_parts):
            str_value = get_valid_label_part(str(value))
            value_label = f'{attr_name}_{i} = {str_value}'
            label_parts.append(value_label)
        return label_parts


def get_valid_label_part(label: str) -> str:
    for replace_char in ESCAPE_CHARS:
        label = label.replace(replace_char, '\\' + replace_char)
    return label


def get_multirow_label_from_parts(parts: Sequence[str]) -> str:
    label = ' | '.join(parts)
    label = '{ ' + label + ' }'
    return label
